"""Telegram configuration."""

from django.conf import settings as django_settings

DEFAULTS = {
    "ROOT_URL": "telegram/",
    "WEBHOOK_URL": "webhook",
    "WEBHOOK_TOKEN": "",
    "ALLOW_SETTINGS_CREATION_FROM_UPDATES": False,
    "REGISTER_DEFAULT_ADMIN": True,
}
REQUIRED = ["BOT_URL"]


class AppSettings:
    """Telegram app settings."""

    def __init__(self):
        """Initialize the settings."""
        app_settings = getattr(django_settings, "TELEGRAM", {}) or {}
        merged_settings = {**DEFAULTS, **app_settings}
        self._settings = merged_settings

    def __getattr__(self, name):
        """Get a setting by name."""
        return self._settings[name]

    def missing_settings(self):
        """Return a list of missing required settings."""
        return [k for k in REQUIRED if k not in self._settings]


settings = AppSettings()



"""Django command to start the bot's poll command."""

from django_telegram_app.management.base import BaseManagementCommand
from tests.testapps.samplebot.telegrambot.commands.poll import Command as PollCommand


class Command(BaseManagementCommand):
    """Start the customcommand command."""

    help = "Start the customcommand command to ask users yes or no."
    command = PollCommand



"""Tests for the get_telegram_settings function."""

from unittest.mock import patch

from django.apps import apps as django_apps
from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase, override_settings

from django_telegram_app import get_telegram_settings_model


class GetModelTests(TestCase):
    """Tests for the get_telegram_settings function."""

    def test_get_telegram_settings_model_default(self):
        """Test that the default TelegramSettings model is returned when no custom model is set."""
        from django_telegram_app.models import TelegramSettings

        model = get_telegram_settings_model()
        self.assertEqual(model, TelegramSettings)

    @override_settings(TELEGRAM_SETTINGS_MODEL="tests_samplebot.CustomTelegramSettings")
    def test_get_telegram_settings_model_custom_model(self):
        """Test that a custom TelegramSettings model is returned when set."""
        from tests.testapps.samplebot.models import CustomTelegramSettings

        model = get_telegram_settings_model()
        self.assertEqual(model, CustomTelegramSettings)

    @override_settings(TELEGRAM_SETTINGS_MODEL="tests.DummyTelegramSettings")
    def test_get_telegram_settings_model_uninstalled_model(self):
        """Test that a custom TelegramSettings model is installed before use."""
        msg = "TELEGRAM_SETTINGS_MODEL refers to model 'tests.CustomTelegramSettings' that has not been installed"
        with self.assertRaises(ImproperlyConfigured, msg=msg):
            get_telegram_settings_model()

    @override_settings(TELEGRAM_SETTINGS_MODEL="invalid_format_model")
    def test_get_telegram_settings_model_invalid_format(self):
        """Test that an improperly formatted TELEGRAM_SETTINGS_MODEL raises an error."""
        msg = "TELEGRAM_SETTINGS_MODEL must be of the form 'app_label.model_name'"
        with self.assertRaises(ImproperlyConfigured, msg=msg):
            get_telegram_settings_model()

    @override_settings(TELEGRAM_SETTINGS_MODEL="tests_samplebot.CustomTelegramSettings")
    def test_get_telegram_settings_model_propagates_unhandled_exceptions(self):
        """Test that unhandled exceptions are propagated."""
        with patch.object(django_apps, "get_model", side_effect=RuntimeError("Unexpected error")):
            with self.assertRaises(RuntimeError, msg="Unexpected error"):
                get_telegram_settings_model()



"""Checks for the telegram app."""

from django.core.checks import Error, register
from django.core.exceptions import ImproperlyConfigured

from django_telegram_app import get_telegram_settings_model
from django_telegram_app.bot import get_commands
from django_telegram_app.conf import settings
from django_telegram_app.models import AbstractTelegramSettings


@register()
def check_telegram_required_settings(app_configs, **kwargs):  # noqa: ARG001  # pylint: disable=unused-argument
    """Check that all required TELEGRAM settings are present."""
    errors = []
    missing = settings.missing_settings()
    if missing:
        errors.append(
            Error(
                f"Missing required TELEGRAM settings: {', '.join(missing)}",
                hint="Add the missing settings in settings.py under the TELEGRAM key.",
                id="telegram.E001",
            )
        )
    return errors


@register()
def check_swappable_telegram_settings(app_configs, **kwargs):  # noqa: ARG001  # pylint: disable=unused-argument
    """Check that TELEGRAM_SETTINGS_MODEL resolves correctly and is a subclass of AbstractTelegramSettings."""
    errors = []

    try:
        model = get_telegram_settings_model()
    except ImproperlyConfigured as exc:
        errors.append(
            Error(
                str(exc),
                hint="Ensure TELEGRAM_SETTINGS_MODEL is defined correctly or omit it to use the default.",
                id="telegram.E002",
            )
        )
        return errors
    except Exception as exc:  # just in case of unexpected errors
        errors.append(
            Error(
                f"Unexpected error while resolving TELEGRAM_SETTINGS_MODEL: {exc!r}",
                id="telegram.E003",
            )
        )
        return errors

    # If resolution succeeded, ensure it subclasses AbstractTelegramSettings
    if not issubclass(model, AbstractTelegramSettings):
        errors.append(
            Error(
                f"{model._meta.label} must subclass telegram.models.base.AbstractTelegramSettings.",
                hint="Update your custom model to inherit from the abstract base.",
                id="telegram.E004",
            )
        )

    return errors


@register()
def check_get_commands(app_configs, **kwargs):  # noqa: ARG001  # pylint: disable=unused-argument
    """Check that get_commands() can load all command classes without errors.

    Since get_commands() is cached, this check also serves to warm the cache.
    """
    errors = []
    try:
        get_commands()
    except Exception as exc:
        errors.append(
            Error(
                f"Error loading commands: {exc!r}",
                hint="Ensure the command modules are defined correctly.",
                id="telegram.E005",
            )
        )
    return errors



"""Django command to set a telegram webhook."""

import requests
from django.core.management.base import BaseCommand, CommandError

from django_telegram_app.conf import settings as app_settings


class Command(BaseCommand):
    """Set a telegram webhook."""

    help = "Sets the webhook for the telegram bot."

    def add_arguments(self, parser):
        """Add command arguments."""
        parser.add_argument(
            "base_url",
            help="A publicly accessible base URL used to construct the Telegram webhook (e.g. 'https://example.com').",
        )

    def handle(self, *_args, **options):
        """Set a webhook.

        Notes: Once a webhook is set, getUpdates no longer works.

        References: https://core.telegram.org/bots/api#setwebhook
        """
        root_url = app_settings.BOT_URL.rstrip("/")
        endpoint = f"{root_url}/setWebhook"
        parts = [options["base_url"], app_settings.ROOT_URL, app_settings.WEBHOOK_URL]
        url = "/".join(part.strip("/") for part in parts if part)
        args = {"url": url}
        if app_settings.WEBHOOK_TOKEN:
            args["secret_token"] = app_settings.WEBHOOK_TOKEN
        response = requests.post(endpoint, json=args, timeout=5)
        response_json: dict = response.json()
        if not response_json.get("ok"):
            self.stderr.write(self.style.ERROR(f"Something went wrong while setting the webhook. {response_json}"))
            raise CommandError(f"Failed to set webhook to {url}.")
        self.stdout.write(self.style.SUCCESS(f'Successfully set webhook to "{url}"'))



# Generated by Django 5.2.5 on 2025-11-09 20:49

import django.core.serializers.json
import django.db.models.deletion
import uuid
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Message',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('raw_message', models.JSONField(verbose_name='raw message')),
                ('error', models.TextField(blank=True, null=True, verbose_name='error')),
            ],
            options={
                'verbose_name': 'message',
                'verbose_name_plural': 'messages',
            },
        ),
        migrations.CreateModel(
            name='TelegramSettings',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('chat_id', models.IntegerField(unique=True, verbose_name='chat id')),
                ('data', models.JSONField(blank=True, default=dict, encoder=django.core.serializers.json.DjangoJSONEncoder, verbose_name='data')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='updated at')),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='user')),
            ],
            options={
                'verbose_name': 'telegram setting',
                'verbose_name_plural': 'telegram settings',
                'swappable': 'TELEGRAM_SETTINGS_MODEL',
            },
        ),
        migrations.CreateModel(
            name='CallbackData',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('token', models.UUIDField(db_index=True, default=uuid.uuid4, unique=True, verbose_name='token')),
                ('command', models.CharField(max_length=255, verbose_name='command')),
                ('step', models.CharField(max_length=255, verbose_name='step')),
                ('action', models.CharField(help_text='Name of a function on the command', max_length=99, verbose_name='action')),
                ('data', models.JSONField(default=dict, encoder=django.core.serializers.json.DjangoJSONEncoder, verbose_name='callback data')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
            ],
            options={
                'verbose_name': 'callback data',
                'verbose_name_plural': 'callback data',
                'indexes': [models.Index(models.F('data__correlation_key'), name='callback_correlation_key_idx')],
            },
        ),
    ]



"""Models for the sample bot app."""

from django.db import models

from django_telegram_app.models import AbstractTelegramSettings


class CustomTelegramSettings(AbstractTelegramSettings):
    """Custom Telegram settings model for testing."""

    extra_field = models.CharField(max_length=100, default="")



"""Tests for admin.py."""

import importlib
from unittest.mock import MagicMock, patch

from django.contrib.admin.options import InlineModelAdmin
from django.test import SimpleTestCase

from django_telegram_app import admin


class AdminTests(SimpleTestCase):
    """Tests for django_telegram_app admin module."""

    def test_inline_is_subclass(self):
        """Test that TelegramSettingInline is a subclass of InlineModelAdmin."""
        assert issubclass(admin.TelegramSettingInline, InlineModelAdmin)

    def test_callbackdata_admin_permissions(self):
        """Test that CallbackDataAdmin permissions are set correctly."""
        from django.contrib.admin.sites import AdminSite

        from django_telegram_app.admin import CallbackDataAdmin
        from django_telegram_app.models import CallbackData

        admin_instance = CallbackDataAdmin(CallbackData, AdminSite("test site"))
        request = MagicMock()  # Mock request object

        assert not admin_instance.has_add_permission(request)
        assert not admin_instance.has_delete_permission(request)
        assert not admin_instance.has_change_permission(request)

    def test_message_admin_permissions(self):
        """Test that MessageAdmin permissions are set correctly."""
        from django.contrib.admin.sites import AdminSite

        from django_telegram_app.admin import MessageAdmin
        from django_telegram_app.models import Message

        admin_instance = MessageAdmin(Message, AdminSite("test site"))
        request = MagicMock()  # Mock request object

        assert not admin_instance.has_add_permission(request)
        assert not admin_instance.has_delete_permission(request)
        assert not admin_instance.has_change_permission(request)

    def test_telegramsettings_admin_not_registered(self):
        """Test that TelegramSettingsAdmin is not registered if the setting REGISTER_DEFAULT_ADMIN is False."""
        from django.contrib.admin.sites import site

        from django_telegram_app.conf import settings

        site._registry.clear()
        with patch.object(settings, "REGISTER_DEFAULT_ADMIN", False):
            importlib.reload(admin)

            assert "TelegramSettings" not in [model.__name__ for model in site._registry.keys()]



"""Telegram admin."""

from django.contrib import admin

from django_telegram_app import get_telegram_settings_model
from django_telegram_app.conf import settings
from django_telegram_app.models import CallbackData, Message

TelegramSettingModel = get_telegram_settings_model()


class TelegramSettingInline(admin.TabularInline):
    """Represent a telegram setting inline in the admin."""

    model = TelegramSettingModel


class TelegramSettingsAdmin(admin.ModelAdmin):
    """Represent the TelegramSettings admin."""


class CallbackDataAdmin(admin.ModelAdmin):
    """Represent the CallbackData admin."""

    list_display = ("token", "data_truncated")

    def has_add_permission(self, request):  # noqa: ARG002  # pylint: disable=unused-argument
        """Do not allow to add callback_data."""
        return False

    def has_delete_permission(self, request, obj=None):  # noqa: ARG002  # pylint: disable=unused-argument
        """Do not allow to delete callback_data."""
        return False

    def has_change_permission(self, request, obj=None):  # noqa: ARG002  # pylint: disable=unused-argument
        """Do not allow to change callback_data."""
        return False


class MessageAdmin(admin.ModelAdmin):
    """Represent the Message admin."""

    list_display = ("update_id", "message_truncated", "error")

    def has_add_permission(self, request):  # noqa: ARG002  # pylint: disable=unused-argument
        """Do not allow to add messages."""
        return False

    def has_delete_permission(self, request, obj=None):  # noqa: ARG002  # pylint: disable=unused-argument
        """Do not allow to delete messages."""
        return False

    def has_change_permission(self, request, obj=None):  # noqa: ARG002  # pylint: disable=unused-argument
        """Do not allow to change messages."""
        return False


admin.site.register(CallbackData, CallbackDataAdmin)
admin.site.register(Message, MessageAdmin)
if settings.REGISTER_DEFAULT_ADMIN:
    admin.site.register(TelegramSettingModel, TelegramSettingsAdmin)



"""Telegram views."""

import json
import logging

from django.contrib.auth.decorators import login_not_required  # type: ignore[reportAttributeAccessIssue]
from django.http import HttpRequest, JsonResponse
from django.views.decorators.csrf import csrf_exempt

from django_telegram_app import models
from django_telegram_app.bot import bot


@csrf_exempt
@login_not_required
def webhook(request: HttpRequest):
    """Handle incoming messages."""
    if not bot.is_valid_token(request.headers.get("X-Telegram-Bot-Api-Secret-Token")):
        return JsonResponse({"status": "error", "message": "Invalid token."}, status=403)
    update = json.loads(request.body)
    message = models.Message(raw_message=update)
    status = "ok"
    try:
        bot.handle_update(update)
    except Exception as exc:
        message.error = str(exc)
        status = "error"
        logging.exception("Error handling Telegram update")
    finally:
        message.save()
    return JsonResponse({"status": status, "message": "Message received."})



"""Echo command for the sample bot."""

from django_telegram_app.bot import bot
from django_telegram_app.bot.base import BaseBotCommand, Step, TelegramUpdate


class Command(BaseBotCommand):
    """Echo command."""

    description = "Responds with the same message."

    @property
    def steps(self):
        """Return the steps of the command."""
        return [WaitForInput(self), Echo(self)]


class WaitForInput(Step):
    """Wait for input step."""

    def handle(self, telegram_update: TelegramUpdate):
        """Handle the step."""
        self.add_waiting_for("userinput")
        bot.send_message(
            "Send the message you want to echo:", self.command.settings.chat_id, message_id=telegram_update.message_id
        )


class Echo(Step):
    """Echo step."""

    def handle(self, telegram_update: TelegramUpdate):
        """Handle the step."""
        data = self.get_callback_data(telegram_update)
        user_input = data["userinput"]
        bot.send_message(
            f"You said: {user_input}",
            self.command.settings.chat_id,
            message_id=telegram_update.message_id,
        )
        self.command.next_step(self.name, telegram_update)



# Generated by Django 5.2.8 on 2025-11-18 19:08

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_telegram_app', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='telegramsettings',
            name='user',
        ),
    ]



"""Telegram bot core module."""

from __future__ import annotations

from typing import TYPE_CHECKING

import requests

from django_telegram_app import get_telegram_settings_model
from django_telegram_app.bot import get_commands, load_command_class
from django_telegram_app.bot.base import TelegramUpdate
from django_telegram_app.conf import settings
from django_telegram_app.models import CallbackData

if TYPE_CHECKING:
    from django_telegram_app.models import AbstractTelegramSettings

DO_NOTHING = "noop"


def is_valid_token(token: str | None):
    """Return whether the webhook token is valid.

    If no token is configured, the token is considered valid.

    Note:
        This token is not the CallbackData token but a token used to validate the webhook.
        This enables us to differentiate original telegram requests from spam
    """
    if not settings.WEBHOOK_TOKEN:
        return True
    return token == settings.WEBHOOK_TOKEN


def handle_update(update: dict, telegram_settings: AbstractTelegramSettings | None = None):
    """Handle the update."""
    telegram_update = TelegramUpdate(update)
    telegram_settings = _get_or_create_telegram_settings(telegram_update, telegram_settings)

    if telegram_update.is_command():
        _start_command_or_send_help(telegram_update, telegram_settings)
    elif telegram_update.is_callback_query():
        _call_command_step(telegram_update.callback_data, telegram_settings, telegram_update)
    elif telegram_settings.data.get("_waiting_for"):
        token = telegram_settings.data["_waiting_for"]
        _call_command_step(token, telegram_settings, telegram_update)
    else:
        send_help(telegram_update.chat_id, telegram_settings)


def send_help(chat_id: int, telegram_settings: "AbstractTelegramSettings"):
    """Send a help message to the user."""
    command_info_list = []
    for command_name, app_name in get_commands().items():
        command = load_command_class(app_name, command_name, telegram_settings)
        command_info_list.append(f"{command.get_command_string()} - {command.description}")

    commands_text = "\n".join(command_info_list)
    help_text = f"Currently available commands:\n{commands_text}"
    send_message(help_text, chat_id)


def send_message(text: str, chat_id: int, reply_markup: dict | None = None, message_id: int = 0):
    """Send a message to the user.

    If message_id is provided, it will edit the existing message instead.

    References:
    https://core.telegram.org/bots/api#sendmessage
    """
    payload = {"chat_id": chat_id, "text": text}
    endpoint = "sendMessage"
    if message_id:
        payload["message_id"] = message_id
        endpoint = "editMessageText"

    if reply_markup:
        payload["reply_markup"] = reply_markup
    post(endpoint, payload=payload)


def post(endpoint: str, payload: dict, timeout: int = 5):
    """Post the payload to the given endpoint."""
    url = _construct_endpoint(endpoint)
    response = requests.post(url, json=payload, timeout=timeout)
    response.raise_for_status()
    return response


@staticmethod
def _construct_endpoint(name: str):
    """Construct the endpoint for the given command."""
    root_url = settings.BOT_URL.rstrip("/")
    return f"{root_url}/{name}"


def _start_command_or_send_help(telegram_update: TelegramUpdate, telegram_settings: "AbstractTelegramSettings"):
    """Start a command or send help message."""
    command_name = telegram_update.message_text.split(maxsplit=1)[0]
    command_str = command_name.lstrip("/")
    try:
        app_name = get_commands()[command_str]
    except KeyError:
        send_help(telegram_update.chat_id, telegram_settings)
        return
    command_cls = load_command_class(app_name, command_str, telegram_settings)
    command_cls.start(telegram_update)


def _call_command_step(token: str, telegram_settings: "AbstractTelegramSettings", telegram_update: TelegramUpdate):
    """Call a command's step from the provided data.

    Return True if the step was called successfully, False otherwise.
    """
    if token == DO_NOTHING:
        return False

    try:
        data = CallbackData.objects.get(token=token)
    except CallbackData.DoesNotExist:
        send_message("This command has expired.", telegram_update.chat_id, message_id=telegram_update.message_id)
        return False

    command_name = data.command.lstrip("/")
    command_info = get_commands()[command_name]
    command = load_command_class(command_info, command_name, telegram_settings)
    getattr(command, data.action)(data.step, telegram_update)
    return True


def _get_or_create_telegram_settings(
    telegram_update: TelegramUpdate, telegram_settings: AbstractTelegramSettings | None = None
):
    """Get or create telegram settings for the given update.

    If telegram_settings is provided, it is returned as is.
    If not provided, attempt to retrieve a TelegramSettings object using the chat_id found in the update.
    If no matching instance exists and `ALLOW_SETTINGS_CREATION_FROM_UPDATES` is True, a new settings instance
    is created by calling TelegramSettingsModel.create_from_telegram_update.
    Otherwise, a DoesNotExist exception is raised.
    """
    if telegram_settings:
        return telegram_settings

    TelegramSettingsModel = get_telegram_settings_model()
    try:
        return TelegramSettingsModel.objects.get(chat_id=telegram_update.chat_id)
    except TelegramSettingsModel.DoesNotExist as exc:
        if not settings.ALLOW_SETTINGS_CREATION_FROM_UPDATES:
            raise exc
        return TelegramSettingsModel.create_from_telegram_update(telegram_update)
